-- # import nat
-- # import digit
-- # import prelude
-- # import tree23
-- # import between
-- # import fingerTree
-- # import maybe
-- # import thunk
-- # import debitPassing
-- # import thunkedVec
-- # import thunkedMaybe
-- # import thunkedSum

infixr 5 ◃, ◃′, ◃◃
infixl 5 ▹, ▹′, ▹▹
infix  5 ▹◃

------------------------------------------------------------------------
-- Cons

mutual
  (◃′) (a : Set) (above : ℕ)
    :  Tree23 a above
    -> FingerTree' a above
    -> FTThunk cons'Steps 1 (FingerTree' a above)
  t ◃′ Empty      = return' $ Single t
  t ◃′ (Single y) = return' $ Deep l r (return Empty)
    where l = digitFT $ digit¹₁ t
          r = digitFT $ digit¹₁ y
  t ◃′ Deep (digitFT l) r m = step $ slower' $
    consHelper l.lenOK t l.elements r m

  consHelper (safeR : IsSafe) (len, above : ℕ) (a : Set)
    :  between 1 len 4
    -> Tree23 a above
    -> Vec len (Tree23 a above)
    -> DigitFT safeR (Tree23 a above)
    -> Thunk (toSteps' len + toSteps safeR)
             (FingerTree' a (above + 1))
    -> Thunk cons'HelperSteps (FingerTree' a above)
  consHelper _ y₁ [y₂] r m =
    return' $ Deep l r (slower' {safeSteps} m)
    where l = digitFT $ digit¹₂ y₁ y₂
  consHelper _ y₁ [y₂, y₃] r m = return' $ Deep l r m
    where l = digitFT $ digit¹₃ y₁ y₂ y₃
  consHelper _ y₁ [y₂, y₃, y₄] r m =
    return' (Deep l r) `ap` pay safeSteps m
    where l = digitFT $ digit¹₄ y₁ y₂ y₃ y₄
  -- Pattern matching on Safe/Unsafe and code duplication to avoid
  -- proof code.
  consHelper {Safe} _ y₁ [y₂, y₃, y₄, y₅] r m = step $
    m `seqBind` \forcedM -> thunkM.do
    m' <- pay (cons'Steps ∸ toSteps Safe) $ Succ3 y₃ y₄ y₅ ◃′ forcedM
    return (Deep (digitFT $ digit¹₂ y₁ y₂) r m')
  consHelper {Unsafe} _ y₁ [y₂, y₃, y₄, y₅] r m = step $
    m `seqBind` \forcedM -> thunkM.do
    m' <- pay (cons'Steps ∸ toSteps Unsafe) $ Succ3 y₃ y₄ y₅ ◃′ forcedM
    return (Deep (digitFT $ digit¹₂ y₁ y₂) r m')
  consHelper {_} {suc (suc (suc (suc _)))}
             (bSS (bZS (bZS (bZS ())))) _ _ _ _

(◃) (a : Set) : a -> FingerTree a -> Thunk consSteps (FingerTree a)
x ◃ t = Zero x ◃′ t

------------------------------------------------------------------------
-- Snoc

mutual
  (▹′) (a : Set) (above : ℕ)
    :  FingerTree' a above
    -> Tree23 a above
    -> FTThunk cons'Steps 1 (FingerTree' a above)
  Empty    ▹′ t = return' $ Single t
  Single y ▹′ t = return' $ Deep l r (return Empty)
    where l = digitFT $ digit¹₁ y
          r = digitFT $ digit¹₁ t
  Deep l (digitFT r) m ▹′ t = step $ slower' $
    snocHelper r.lenOK l r.elements t m

  snocHelper (safeL : IsSafe) (len', above : ℕ) (a : Set)
    :  between 1 len' 4
    -> DigitFT safeL (Tree23 a above)
    -> Vec len' (Tree23 a above)
    -> Tree23 a above
    -> Thunk (toSteps safeL + toSteps' len')
             (FingerTree' a (above + 1))
    -> Thunk cons'HelperSteps (FingerTree' a above)
  snocHelper _ l [y₁] y₂ m =
    return' $ Deep l r (slower' {safeSteps} m)
    where r = digitFT $ digit¹₂ y₁ y₂
  snocHelper _ l [y₁, y₂] y₃ m = return' $ Deep l r m
    where r = digitFT $ digit¹₃ y₁ y₂ y₃
  snocHelper _ l [y₁, y₂, y₃] y₄ m =
    return' (Deep l r) `ap` pay safeSteps m
    where r = digitFT $ digit¹₄ y₁ y₂ y₃ y₄
  snocHelper {Safe} _ l [y₁, y₂, y₃, y₄] y₅ m = step $
    m `seqBind` \forcedM -> thunkM.do
    m' <- pay (cons'Steps ∸ toSteps Safe) $ forcedM ▹′ Succ3 y₁ y₂ y₃
    return (Deep l (digitFT $ digit¹₂ y₄ y₅) m')
  snocHelper {Unsafe} _ l [y₁, y₂, y₃, y₄] y₅ m = step $
    m `seqBind` \forcedM -> thunkM.do
    m' <- pay (cons'Steps ∸ toSteps Unsafe) $ forcedM ▹′ Succ3 y₁ y₂ y₃
    return (Deep l (digitFT $ digit¹₂ y₄ y₅) m')
  snocHelper {_} {suc (suc (suc (suc _)))}
             (bSS (bZS (bZS (bZS ())))) _ _ _ _

(▹) (a : Set) : FingerTree a -> a -> Thunk consSteps (FingerTree a)
t ▹ x = t ▹′ Zero x

exampleSnocNoElements : ℕ
exampleSnocNoElements = 10

exampleSnocSteps : ℕ
exampleSnocSteps = exampleSnocNoElements * (consSteps + 2)

exampleSnoc : Thunk exampleSnocSteps (FingerTree ℕ)
exampleSnoc = tFoldr (flip (▹)) Empty =<< tEnumFrom exampleSnocNoElements 1

------------------------------------------------------------------------
-- Left and right view helpers

digitToFT (a : Set) (above : ℕ) (safe : IsSafe)
  : DigitFT safe (Tree23 a above) -> FingerTree' a above
digitToFT (digitFT d) = helper d.lenOK d.elements
  where
  helper (a : Set) (len, above : ℕ)
    :  between 1 (len + 1) 4
    -> Vec (len + 1) (Tree23 a above)
    -> FingerTree' a above
  helper _ [x₁]             = Single x₁
  helper _ [x₁, x₂]         = Deep (digitFT $ digit¹₁ x₁)
                                   (digitFT $ digit¹₁ x₂)
                                   (return Empty)
  helper _ [x₁, x₂, x₃]     = Deep (digitFT $ digit¹₂ x₁ x₂)
                                   (digitFT $ digit¹₁ x₃)
                                   (return' Empty)
  helper _ [x₁, x₂, x₃, x₄] = Deep (digitFT $ digit¹₂ x₁ x₂)
                                   (digitFT $ digit¹₂ x₃ x₄)
                                   (return' Empty)
  helper {_} {suc (suc (suc (suc _)))}
         (bSS (bZS (bZS (bZS ())))) _

tree23ToDigit (a : Set) (above : ℕ)
  : Tree23 a (above + 1) -> DigitFT Safe (Tree23 a above)
tree23ToDigit (Succ2 l r)   = digitFT $ digit¹₂ l r
tree23ToDigit (Succ3 l m r) = digitFT $ digit¹₃ l m r

------------------------------------------------------------------------
-- Left view

mutual

  viewL' (above : ℕ) (a : Set)
    :  FingerTree' a above
    -> Thunk viewL'Steps
             (Maybe (Tree23 a above
                       ×
                     FTThunk deepLSteps 1 (FingerTree' a above)
                    ))
  viewL' Empty                  = return' Nothing
  viewL' (Single x)             = return' $ Just $ pair x (return' Empty)
  viewL' (Deep (digitFT l) r m) = step $ thunkM.do
    lHead <- headDigit l
    lTail <- tailDigit l
    return $ Just $ pair lHead (deepL r (pair lTail m))

  deepL (len, above : ℕ) (a : Set) (safeR : IsSafe)
    :  DigitFT safeR (Tree23 a above)
    -> Digit 0 len 3 (Tree23 a above)
         ×
       Thunk (toSteps' (len + 1) + toSteps safeR)
             (FingerTree' a (above + 1))
    -> FTThunk deepLSteps 1 (FingerTree' a above)
  deepL {suc len} r (pair l m) = slower' $ thunkM.do
      -- Steps for m: toSteps' (len + 2) + toSteps safeR.
      -- Steps m should take: toSteps' (len + 1) + toSteps safeR.
      -- Pay off difference (assume worst possible case).
      m' <- pay safeSteps m
      return $ Deep (digitFT $ newLimits b l) r $ slower prop m'
    where
    b : between 1 (len + 1) 4
    b with toLE l.lenOK =>
      _ (pair _ le') = fromLE (leS leZ) (le' `transitive` sucIncreasing)

    -- This one should be left to QuickCheck-plugin...
    prop (len, x : ℕ)
      : (toSteps' (len + 2) + x) ∸ safeSteps ≤ toSteps' (len + 1) + x
    prop {0} = reflexive
    prop {1} = plusIncreasingLeft {safeSteps}
    prop {2} = subtractDecreasing $ plusIncreasingLeft {safeSteps}
    prop {3} = subtractDecreasing reflexive
    prop {suc (suc (suc (suc _)))} = subtractDecreasing reflexive

  deepL {0} r (pair _ m) = step $ substThunk prop refl' $
    helper r =<< viewL' =<< m
    where
    prop (safe : IsSafe)
      : toSteps safe + viewL'Steps +
        (1 + (deepLHelperSteps + safeSteps ∸ toSteps safe ∸ 1))
        ≡ deepLSteps + safeSteps ∸ 1
    prop {Unsafe} = refl
    prop {Safe}   = refl

    helper (safe : IsSafe) (above : ℕ) (a : Set)
      :  DigitFT safe (Tree23 a above)
      -> Maybe (Tree23 a (above + 1)
                  ×
                FTThunk deepLSteps 1 (FingerTree' a (above + 1))
               )
      -> Thunk (1 + (deepLHelperSteps + safeSteps ∸ toSteps safe ∸ 1))
               (FingerTree' a above)
    helper r' Nothing = return' $ digitToFT r'
    helper {safe} r' (Just (pair mHead mTailThunk)) = step $
      slower prop $ thunkM.do
        mTail <- pay (safeSteps ∸ toSteps safe) mTailThunk
        return $ Deep (tree23ToDigit mHead) r' $ slower le mTail
      where
      le (safe' : IsSafe)
        : (deepLSteps + safeSteps)
          ∸ (safeSteps ∸ toSteps safe')
          ≤ safeSteps + toSteps safe'
      -- Requirement: deepLSteps ≤ safeSteps.
      le {Safe}   = plusIncreasingRight
      le {Unsafe} = plusIncreasingRight

      prop (safe : IsSafe)
        : safeSteps ∸ toSteps safe
          ≤ deepLHelperSteps + safeSteps ∸ toSteps safe ∸ 1
      prop {Safe}   = plusIncreasingRight
      prop {Unsafe} = plusIncreasingRight

-- Returns Nothing for empty trees.

viewL (a : Set)
  : FingerTree a -> Thunk viewLSteps (Maybe (a × FingerTree a))
viewL t = tMapMaybe fix =<< viewL' t
  where
  fix (a : Set)
    :  Tree23 a 0 × FTThunk deepLSteps 1 (FingerTree a)
    -> FTThunk deepLSteps 1 (a × FingerTree a)
  fix (pair (Zero h) tT) = tPair h =<< tT

exampleViewLSteps : ℕ
exampleViewLSteps = exampleSnocSteps + viewLSteps

exampleViewL : Thunk exampleViewLSteps (Maybe (ℕ × FingerTree ℕ))
exampleViewL = viewL =<< exampleSnoc

------------------------------------------------------------------------
-- Right view

mutual

  viewR' (above : ℕ) (a : Set)
    :  FingerTree' a above
    -> Thunk viewR'Steps
             (Maybe (FTThunk deepRSteps 1 (FingerTree' a above)
                       ×
                     Tree23 a above
                    ))
  viewR' Empty                  = return' Nothing
  viewR' (Single x)             = return' $ Just $ pair (return' Empty) x
  viewR' (Deep l (digitFT r) m) = step $ thunkM.do
    rLast <- lastDigit r
    rInit <- initDigit r
    return $ Just $ pair (deepR l $ pair rInit m) rLast

  deepR (len, above : ℕ) (a : Set) (safeL : IsSafe)
    :  DigitFT safeL (Tree23 a above)
    -> Digit 0 len 3 (Tree23 a above)
         ×
       Thunk (toSteps safeL + toSteps' (len + 1))
             (FingerTree' a (above + 1))
    -> FTThunk deepRSteps 1 (FingerTree' a above)
  deepR {suc len} l (pair r m) = slower' $ thunkM.do
      m' <- pay safeSteps m
      return $ Deep l (digitFT (newLimits b r)) $ slower prop m'
    where
    b : between 1 (len + 1) 4
    b with toLE r.lenOK =>
      _ (pair _ le') = fromLE (leS leZ) (le' `transitive` sucIncreasing)

    -- This one should be left to QuickCheck-plugin...
    prop (len, x : ℕ)
      : (x + toSteps' (len + 2)) ∸ safeSteps ≤ x + toSteps' (len + 1)
    prop {0} = reflexive
    prop {1} = plusIncreasingLeft {safeSteps}
    prop {2} = subtractDecreasing $ plusIncreasingLeft {safeSteps}
    prop {3} = subtractDecreasing reflexive
    prop {suc (suc (suc (suc _)))} = subtractDecreasing reflexive

  deepR {0} l (pair _ m) = step $ substThunk prop refl' $
    helper l =<< viewR' =<< m
    where
    prop (safe : IsSafe)
      : toSteps safe + viewR'Steps +
        (deepLHelperSteps + safeSteps ∸ toSteps safe)
        ≡ deepRSteps + safeSteps ∸ 1
    prop {Unsafe} = refl
    prop {Safe}   = refl

    helper (safe : IsSafe) (above : ℕ) (a : Set)
      :  DigitFT safe (Tree23 a above)
      -> Maybe (FTThunk deepRSteps 1 (FingerTree' a (above + 1))
                  ×
                Tree23 a (above + 1)
               )
      -> Thunk (deepLHelperSteps + safeSteps ∸ toSteps safe)
               (FingerTree' a above)
    -- By pattern matching on Safe/Unsafe below we get rid of _three_
    -- separate (but easy) proofs. The code is clearly shorter, even
    -- though it is now duplicated. See deepL for the cleaner but
    -- longer version of this code.

    -- The following alternatives all seem to be more appealing:

    -- helper {safe} l' Nothing = case-plugin(safe, case $ digitToFT l')
    -- helper l' Nothing = slower agsy-plugin() $ case {0} $ digitToFT r'
    -- helper {safe} l' Nothing =
    --   slower QuickCheck-plugin(safe, arbitraryIsSafe) $
    --     case {0} $ digitToFT r'

    helper {Unsafe} l' Nothing = return' $ digitToFT l'
    helper {Safe}   l' Nothing = return' $ digitToFT l'
    helper {Unsafe} l' (Just (pair mInitThunk mLast)) = step $ thunkM.do
        mInit <- pay (safeSteps ∸ toSteps Unsafe) mInitThunk
        return $ Deep l' (tree23ToDigit mLast) $ slower' mInit
    helper {Safe}   l' (Just (pair mInitThunk mLast)) = step $ thunkM.do
        mInit <- pay (safeSteps ∸ toSteps Safe) mInitThunk
        return $ Deep l' (tree23ToDigit mLast) $ slower' mInit

viewR (a : Set)
  :  FingerTree a
  -> Thunk viewRSteps (Maybe (FingerTree a × a))
viewR t = tMapMaybe fix =<< viewR' t
  where
  fix (a : Set)
    :  FTThunk deepRSteps 1 (FingerTree a) × Tree23 a 0
    -> FTThunk deepRSteps 1 (FingerTree a × a)
  fix (pair iT (Zero l)) = thunkM.do
    i <- iT
    tPair i l

exampleViewRSteps : ℕ
exampleViewRSteps = exampleSnocSteps + viewRSteps

exampleViewR : Thunk exampleViewRSteps (Maybe (FingerTree ℕ × ℕ))
exampleViewR = viewR =<< exampleSnoc

------------------------------------------------------------------------
-- Append

-- The append function illustrates the use of a dependent bind. The
-- new library does not contain a dependent bind, so I refrained from
-- mentioning the append function in the paper.

digitConsSteps : ℕ
digitConsSteps = 4 * (cons'Steps + safeSteps + 1)

(▹▹) (a : Set) (above : ℕ) (safe : IsSafe)
  :  FingerTree' a above -> DigitFT safe (Tree23 a above)
  -> Thunk digitConsSteps (FingerTree' a above)
e ▹▹ digitFT d = slower le $ tFoldr (flip (▹′)) e d.elements
  where
  le = mulMon (snd $ toLE d.lenOK)
              (reflexive {cons'Steps + safeSteps + 1})

(◃◃) (a : Set) (above : ℕ) (safe : IsSafe)
  :  DigitFT safe (Tree23 a above) -> FingerTree' a above
  -> Thunk digitConsSteps (FingerTree' a above)
digitFT d ◃◃ e = slower le $ tFoldr (◃′) e d.elements
  where
  le = mulMon (snd $ toLE d.lenOK)
              (reflexive {cons'Steps + safeSteps + 1})

predLenAfterNodes : ℕ -> ℕ
predLenAfterNodes 0 = 0
predLenAfterNodes 1 = 0
predLenAfterNodes 2 = 1
predLenAfterNodes (suc (suc (suc n))) = 1 + predLenAfterNodes n

safeAfterNodes : ℕ -> IsSafe
safeAfterNodes len = isSafe (predLenAfterNodes len + 1)

nodesHelperSteps : ℕ -> ℕ
nodesHelperSteps n = n * (1 + consDigitSteps)

nodesSteps : ℕ
nodesSteps = nodesHelperSteps 4

nodes (n, len, above : ℕ) (a : Set)
  :  (d : Digit 2 (len + 2) 12 (Tree23 a above))
  -> Thunk nodesSteps (DigitFT (safeAfterNodes len)
                               (Tree23 a (above + 1))
                      )
nodes d = return digitFT `ap` helper 4 d.lenOK d.elements
  where
  helper (len, above : ℕ) (a : Set)
    :  (n : ℕ)
    -> between 2 (len + 2) (3 * n)
    -> Vec (len + 2) (Tree23 a above)
    -> Thunk (nodesHelperSteps n)
             (Digit 1 (predLenAfterNodes len + 1) n (Tree23 a (above + 1)))
  helper (suc _) _ [t₁, t₂]     = return' $ digit¹₁ (Succ2 t₁ t₂)
  helper (suc _) _ [t₁, t₂, t₃] = return' $ digit¹₁ (Succ3 t₁ t₂ t₃)
  helper (suc (suc _)) _ [t₁, t₂, t₃, t₄] = return' $
    digit¹₂ (Succ2 t₁ t₂) (Succ2 t₃ t₄)
  helper (suc (suc n)) (bSS (bSS (bZS (bZS (bZS b)))))
         (t₁ :: t₂ :: t₃ :: t₄ :: t₅ :: ts) = step $ thunkM.do
    d'  <- helper (n + 1) (bSS $ bSS b) (t₄ :: t₅ :: ts)
    d'' <- consDigit (Succ3 t₁ t₂ t₃) d'
    return $ extendDigit 1 0 d''
  helper (suc zero) (bSS (bSS (bZS ()))) _

app₃Steps : ℕ -> ℕ
app₃Steps d = b + d * s
  where
  b = cons'Steps + safeSteps + digitConsSteps
  s = appendDigitSteps 4 + appendDigitSteps 8 + nodesSteps + 1 + 4 * safeSteps

app₃ (a : Set) (above : ℕ) (safe : IsSafe)
  :  (l : FingerTree' a above)
  -> Maybe (DigitFT safe (Tree23 a above))
  -> (r : FingerTree' a above)
  -> Thunk (app₃Steps (depthFT l `min` depthFT r)) (FingerTree' a above)
app₃ Empty      Nothing  r          = return' r
app₃ Empty      (Just m) r          = slower' $ m ◃◃ r
app₃ l          Nothing  Empty      = return' l
app₃ l          (Just m) Empty      = slower' $ l ▹▹ m
app₃ (Single x) Nothing  r          = slower' $ x ◃′ r
app₃ (Single x) (Just m) r          = (◃′) x =<< m ◃◃ r
app₃ l          Nothing  (Single y) = slower' $ l ▹′ y
app₃ l          (Just m) (Single y) = l ▹▹ m >>= flip (▹′) y
app₃ {a} {above}
     (Deep l₁ (digitFT r₁) m₁)
     Nothing
     (Deep  (digitFT l₂) r₂ m₂) = step $ slower prop $ thunkM.do
    d   <- r₁ `appendDigit` l₂
    mid <- nodes {4} (extendDigit 0 4 d)
    let stepFun : FingerTree' a (above + 1) -> FingerTree' a (above + 1) -> ℕ
        stepFun m₁' m₂' = app₃Steps (depthFT m₁' `min` depthFT m₂')
    m' <- dependentBind₂ stepFun m₁ m₂ (\m₁' m₂' -> app₃ m₁' (Just mid) m₂')
    return $ Deep l₁ r₂ (slower leZ $ return m')
  where
  -- It would be nice if QuickCheck-plugin or Agsy-plugin could handle
  -- this one.
  prop (s₁, s₂, s₃, s₄ : IsSafe)
       (m₁', m₂' : FingerTree' a (above + 1))
    : appendDigitSteps 4 + (nodesSteps +
      ((toSteps s₁ + toSteps s₂) + (toSteps s₃ + toSteps s₄) +
       app₃Steps (depthFT m₁' `min` depthFT m₂')
      ))
      ≤ app₃Steps (1 + (depthFT m₁' `min` depthFT m₂')) ∸ 1
  prop = plusIncreasingRight {appendDigitSteps 8} {appendDigitSteps 4}
           `plusMon`
         (reflexive {nodesSteps}
            `plusMon`
          ((prop_toSteps₂ `plusMon` prop_toSteps₂)
            `plusMon`
           reflexive {app₃Steps (depthFT _ `min` depthFT _)}
          )
         )
app₃ {a} {above}
     (Deep l₁ (digitFT r₁) m₁)
     (Just (digitFT m))
     (Deep (digitFT l₂) r₂ m₂) = step $ slower prop $ thunkM.do
    d   <- r₁ `appendDigit` m >>= flip appendDigit l₂
    mid <- nodes {4} (extendDigit 1 0 d)
    let stepFun : FingerTree' a (above + 1) -> FingerTree' a (above + 1) -> ℕ
        stepFun m₁' m₂' = app₃Steps (depthFT m₁' `min` depthFT m₂')
    m' <- dependentBind₂ stepFun m₁ m₂ (\m₁' m₂' -> app₃ m₁' (Just mid) m₂')
    return $ Deep l₁ r₂ (slower leZ $ return m')
  where
  prop (s₁, s₂, s₃, s₄ : IsSafe)
       (m₁', m₂' : FingerTree' a (above + 1))
    : (appendDigitSteps 4 + appendDigitSteps 8) + (nodesSteps +
      ((toSteps s₁ + toSteps s₂) + (toSteps s₃ + toSteps s₄) +
       app₃Steps (depthFT m₁' `min` depthFT m₂')
      ))
      ≤ app₃Steps (1 + (depthFT m₁' `min` depthFT m₂')) ∸ 1
  prop = reflexive {appendDigitSteps 4 + appendDigitSteps 8}
           `plusMon`
         (reflexive {nodesSteps}
            `plusMon`
          ((prop_toSteps₂ `plusMon` prop_toSteps₂)
            `plusMon`
           reflexive {app₃Steps (depthFT _ `min` depthFT _)}
          )
         )

appendFTSteps : ℕ -> ℕ
appendFTSteps = app₃Steps

(▹◃) (a : Set)
  :  (l, r : FingerTree a)
  -> Thunk (appendFTSteps (depthFT l `min` depthFT r)) (FingerTree a)
t₁ ▹◃ t₂ = app₃ {_} {_} {Safe} t₁ Nothing t₂

exampleAppendFTSteps : ℕ
exampleAppendFTSteps =
  appendFTSteps (depthFT (force exampleSnoc) `min`
                 depthFT (force exampleFT)) +
  exampleSnocSteps

exampleAppendFT : Thunk exampleAppendFTSteps (FingerTree ℕ)
exampleAppendFT = dependentBind₂ stepFun exampleFT exampleSnoc (▹◃)
  where
  stepFun : FingerTree ℕ -> FingerTree ℕ -> ℕ
  stepFun l r = appendFTSteps (depthFT l `min` depthFT r)
