-- # import nat
-- # import maybe
-- # import eq
-- # import sum
-- # import pair

infix 4 ≤, ≥, ≤?, ≤?>, ≤?≥

data (≤) : ℕ -> ℕ -> Set where
  leZ (n : ℕ)    : 0 ≤ n
  leS (m, n : ℕ) : m ≤ n -> suc m ≤ suc n

leUnique (m, n : ℕ) : (le₁, le₂ : m ≤ n) -> le₁ ≡ le₂
leUnique leZ       leZ       = refl
leUnique (leS le₁) (leS le₂) = cong leS (leUnique le₁ le₂)

antisymmetric (i, j : ℕ) : i ≤ j -> j ≤ i -> i ≡ j
antisymmetric leZ       leZ       = refl
antisymmetric (leS le₁) (leS le₂) = cong suc (antisymmetric le₁ le₂)

(≥) : ℕ -> ℕ -> Set
m ≥ n = n ≤ m

(≤?) : (m, n : ℕ) -> Maybe (m ≤ n)
zero   ≤? n      = Just leZ
suc m' ≤? zero   = Nothing
suc m' ≤? suc n' = mapMaybe leS (m' ≤? n')

(≤?>) : (m, n : ℕ) -> m ≤ n ∔ n + 1 ≤ m
zero   ≤?> n      = left leZ
suc m' ≤?> zero   = right $ leS leZ
suc m' ≤?> suc n' = mapSum leS leS (m' ≤?> n')

transitive (i, j, k : ℕ) : i ≤ j -> j ≤ k -> i ≤ k
transitive leZ      _         = leZ
transitive (leS le) (leS le') = leS (transitive le le')

castLE (x, y, x', y' : ℕ)
  : x ≡ y -> x' ≡ y' -> x ≤ x' -> y ≤ y'
castLE refl refl le = le

------------------------------------------------------------------------
-- Not less than or equal

absurdLE (n : ℕ) (whatever : Set) : suc n ≤ 0 -> whatever
absurdLE ()

------------------------------------------------------------------------
-- Monotonicity

monotone : (ℕ -> ℕ) -> Set
monotone f = (i, j : ℕ) -> i ≤ j -> f i ≤ f j

bimonotone : (ℕ -> ℕ -> ℕ) -> Set
bimonotone (+*) =
  (i₁, i₂, j₁, j₂ : ℕ) -> i₁ ≤ j₁ -> i₂ ≤ j₂ -> (i₁ +* i₂) ≤ (j₁ +* j₂)

plusIncreasingLeft (n, m : ℕ) : m ≤ n + m
plusIncreasingLeft {_} {0}     = leZ
plusIncreasingLeft {_} {suc m} = leS plusIncreasingLeft

plusIncreasingRight (n, m : ℕ) : m ≤ m + n
plusIncreasingRight {_} {0}     = leZ
plusIncreasingRight {_} {suc m} = leS plusIncreasingRight

plusLeftMon (n, i, j : ℕ) : i ≤ j -> i + n ≤ j + n
plusLeftMon leZ      = plusIncreasingLeft
plusLeftMon (leS le) = leS (plusLeftMon le)

plusRightMon (n, i, j : ℕ) : i ≤ j -> n + i ≤ n + j
plusRightMon leZ      = plusIncreasingRight
plusRightMon (leS le) = leS (plusRightMon le)

plusMon (i, j, m, n : ℕ) : i ≤ j -> m ≤ n -> i + m ≤ j + n
plusMon le1 le2 = plusLeftMon le1 `transitive` plusRightMon le2

plusMon' : bimonotone (+)
plusMon' _ _ _ _ = plusMon

plusLeftIncreasing (k, i, j : ℕ) : i ≤ j -> i ≤ k + j
plusLeftIncreasing le = transitive le plusIncreasingLeft

plusRightIncreasing (k, i, j : ℕ) : i ≤ j -> i ≤ j + k
plusRightIncreasing le = transitive le plusIncreasingRight

reflexive (n : ℕ) : n ≤ n
reflexive = plusIncreasingLeft {0}

liftEq (x, y : ℕ)
  : x ≡ y -> x ≤ y
liftEq refl = reflexive

biToMonLeft ((+*) : ℕ -> ℕ -> ℕ)
  : (j : ℕ) -> bimonotone (+*) -> monotone (\i -> i +* j)
biToMonLeft j bimon = \i₁ i₂ le -> bimon i₁ j i₂ j le reflexive

biToMonRight ((+*) : ℕ -> ℕ -> ℕ)
  : (j : ℕ) -> bimonotone (+*) -> monotone (\i -> j +* i)
biToMonRight j bimon = \i₁ i₂ le -> bimon j i₁ j i₂ reflexive le

sucIncreasing (n : ℕ) : n ≤ suc n
sucIncreasing = plusLeftIncreasing {1} reflexive

smallerRelated (m, n : ℕ) : n ∸ m ≤ n
smallerRelated {_}     {0}     = leZ
smallerRelated {0}     {_}     = reflexive
smallerRelated {suc m} {suc n} = transitive smallerRelated sucIncreasing

subtractLarger (m, n : ℕ) : m ≤ n -> (m ∸ n) ≡ 0
subtractLarger leZ      = refl
subtractLarger (leS le) = subtractLarger le

subtractDecreasing (k, i, j : ℕ) : i ≤ j -> i ∸ k ≤ j
subtractDecreasing le = transitive smallerRelated le

subtractMonotoneHelper (i, j : ℕ) : i ≤ j -> i ∸ 1 ≤ j ∸ 1
subtractMonotoneHelper leZ      = leZ
subtractMonotoneHelper (leS le) = le

subtractMonotone (k, i, j : ℕ) : i ≤ j -> i ∸ k ≤ j ∸ k
subtractMonotone {0}     le = le
subtractMonotone {suc k} le =
  castLE subtractLinear subtractLinear
         (subtractMonotone {k} (subtractMonotoneHelper le))

{-

    i ≤ j -> i ∸ (k + 1) ≤ j ∸ (k + 1)
<=> { Linearity }
    i ≤ j -> (i ∸ 1) ∸ k ≤ (j ∸ 1) ∸ k
<=  { IH }
    i ≤ j -> i ∸ 1 ≤ j ∸ 1

-}

subtractAntiMonotone (k, j, i : ℕ) : i ≥ j -> k ∸ i ≤ k ∸ j
subtractAntiMonotone         leZ      = subtractDecreasing reflexive
subtractAntiMonotone {0}     (leS le) = leZ
subtractAntiMonotone {suc k} (leS le) = subtractAntiMonotone le

subtractMonotoneAntimonotone (i, j, m, n : ℕ)
  : i ≤ j -> m ≥ n -> i ∸ m ≤ j ∸ n
subtractMonotoneAntimonotone le1 le2 =
  subtractMonotone le1 `transitive` subtractAntiMonotone le2

prop_plus_subtract (m, n : ℕ) : m ≤ n -> m + (n ∸ m) ≡ n
prop_plus_subtract leZ      = refl
prop_plus_subtract (leS le) = cong suc (prop_plus_subtract le)

subtractAddAssoc (k, i, j : ℕ)
  : k ≤ i -> (i ∸ k) + j ≡ (i + j) ∸ k
subtractAddAssoc leZ      = refl
subtractAddAssoc (leS le) = subtractAddAssoc le

subtractAddAssocLE (i, k, j : ℕ)
  : (j + i) ∸ k ≤ i ∸ k + j
subtractAddAssocLE {0}             = subtractDecreasing reflexive
subtractAddAssocLE {suc i} {0}     = liftEq (cong suc addCommutative)
subtractAddAssocLE {suc i} {suc k} = subtractAddAssocLE {i} {k}

dropPlusMinus (j, i, k : ℕ)
  : j ≤ k -> (i + j) ∸ k ≤ i
dropPlusMinus {j} {i} {k} le =
  rewrite toPropEq (subtractLarger le)
  in subtractAddAssocLE {j} {k} {i}

dropPlusRightDecreasing (i, j, k : ℕ) : i + j ≤ k -> i ≤ k
dropPlusRightDecreasing {i} {j} {k} le =
  castLE subtractPlusInverses
         refl
         (subtractDecreasing {j} {i + j} {k} le)

dropPlusLeftDecreasing (i, j, k : ℕ) : j + i ≤ k -> i ≤ k
dropPlusLeftDecreasing {i} {j} {k} le =
  castLE subtractPlusInverses'
         refl
         (subtractDecreasing {j} {j + i} {k} le)

------------------------------------------------------------------------

mulIncreasingRight (n, m : ℕ) : (1 ≤ n) -> m ≤ m * n
mulIncreasingRight {_}     {0}     _  = leZ
mulIncreasingRight {suc n} {suc m} le =
  leS (plusLeftIncreasing (mulIncreasingRight le))

mulIncreasingLeft (n, m : ℕ) : (1 ≤ n) -> m ≤ n * m
mulIncreasingLeft le =
  castLE refl mulCommutative
         (mulIncreasingRight le)

mulMon (i, j, m, n : ℕ) : (i ≤ m) -> (j ≤ n) -> (i * j ≤ m * n)
mulMon leZ        _    = leZ
mulMon (leS iLeM) jLeN = plusMon jLeN (mulMon iLeM jLeN)

------------------------------------------------------------------------

infixl 9 `min`, `max`

min : (m, n : ℕ) -> ℕ
min (suc m) (suc n) = suc (m `min` n)
min (suc _) zero    = zero
min zero    _       = zero

minRightZero (n : ℕ) : n `min` 0 ≡ 0
minRightZero {0}     = refl
minRightZero {suc _} = refl

minRightZero' : (n : ℕ) -> n `min` 0 ≡ 0
minRightZero' _ = toPropEq minRightZero

rewrite minRightZero'

minCommutative (i, j : ℕ) : i `min` j ≡ j `min` i
minCommutative {suc i} {suc j} = cong suc minCommutative
minCommutative {0}     {_}     = refl
minCommutative {_}     {0}     = refl

minMonotone (m₁, m₂, n₁, n₂ : ℕ)
  : m₁ ≤ m₂ -> n₁ ≤ n₂ -> m₁ `min` n₁ ≤ m₂ `min` n₂
minMonotone leZ       _         = leZ
minMonotone (leS _)   leZ       = leZ
minMonotone (leS le₁) (leS le₂) = leS $ minMonotone le₁ le₂

minEitherOr : (x, y : ℕ) -> (x `min` y ≡ x) ∔ (x `min` y ≡ y)
minEitherOr 0       _ = left refl
minEitherOr (suc _) 0 = right refl
minEitherOr (suc x) (suc y) with minEitherOr x y =>
  _ _ _ (left  eq) = left  (cong suc eq)
  _ _ _ (right eq) = right (cong suc eq)

minLELeft (x, y : ℕ) : x ≤ y -> x `min` y ≡ x
minLELeft leZ      = refl
minLELeft (leS le) = cong suc (minLELeft le)

minLERight (x, y : ℕ) : x ≥ y -> x `min` y ≡ y
minLERight le = minCommutative `trans` minLELeft le

minLeft (x, y : ℕ) : x `min` y ≤ x
minLeft {0}             = leZ
minLeft {suc x} {0}     = leZ
minLeft {suc x} {suc y} = leS minLeft

minRight (y, x : ℕ) : x `min` y ≤ y
minRight = liftEq minCommutative `transitive` minLeft

leMinDistrib₁ (i, m, n : ℕ) : i ≤ m `min` n -> i ≤ m × i ≤ n
leMinDistrib₁ leZ = pair leZ leZ
leMinDistrib₁ {_} {suc _} {suc _} (leS le) with leMinDistrib₁ le =>
  _ _ (pair le₁ le₂) = pair (leS le₁) (leS le₂)

leMinDistrib₂ (i, m, n : ℕ) : i ≤ m -> i ≤ n -> i ≤ m `min` n
leMinDistrib₂ leZ le = leZ
leMinDistrib₂ (leS le₁) (leS le₂) with leMinDistrib₂ le₁ le₂ =>
  _ _ _ le = leS le

leMinDistrib (i, m, n : ℕ) : i ≤ m `min` n ⇔ (i ≤ m × i ≤ n)
leMinDistrib = pair leMinDistrib₁ (uncurry leMinDistrib₂)

homomorphicMin (f : ℕ -> ℕ)
  : monotone f -> (m, n : ℕ) -> f (m `min` n) ≡ f m `min` f n
homomorphicMin mon m n with minEitherOr m n =>
  _ mon m n (left  eqM) = rewrite-plugin(eqM,
    sym $ minLELeft $ mon m n (liftEq (sym eqM) `transitive` minRight))
  _ mon m n (right eqN) = rewrite-plugin(eqN,
    sym $ minLERight $ mon n m (liftEq (sym eqN) `transitive` minLeft))

bihomomorphicMin ((+*) : ℕ -> ℕ -> ℕ)
  :  bimonotone (+*) -> (m₁, m₂, n₁, n₂ : ℕ)
  -> (m₁ `min` m₂) +* (n₁ `min` n₂) ≡
     ((m₁ +* n₁) `min` (m₁ +* n₂)) `min` ((m₂ +* n₁) `min` (m₂ +* n₂))
bihomomorphicMin bimon m₁ m₂ n₁ n₂ =
  homomorphicMin (biToMonLeft (n₁ `min` n₂) bimon) m₁ m₂
    `trans`
  congBin min (homomorphicMin (biToMonRight m₁ bimon) n₁ n₂)
              (homomorphicMin (biToMonRight m₂ bimon) n₁ n₂)

bihomomorphicMin' ((+*) : ℕ -> ℕ -> ℕ)
  :  bimonotone (+*) -> (m₁, m₂, n₁, n₂ : ℕ)
  -> (m₁ `min` m₂) +* (n₁ `min` n₂) ≤ (m₁ +* n₁) `min` (m₂ +* n₂)
bihomomorphicMin' bimon _ _ _ _ =
  liftEq (bihomomorphicMin bimon _ _ _ _)
    `transitive`
  minMonotone minLeft minRight

max : (m, n : ℕ) -> ℕ
max (suc m) (suc n) = suc (m `max` n)
max (suc n) zero    = suc n
max zero    n       = n

maxRightZero (n : ℕ) : n `max` 0 ≡ n
maxRightZero {0}     = refl
maxRightZero {suc _} = refl

maxRightZero' : (n : ℕ) -> n `max` 0 ≡ n
maxRightZero' _ = toPropEq maxRightZero

rewrite maxRightZero'

maxCommutative (i, j : ℕ) : i `max` j ≡ j `max` i
maxCommutative {suc i} {suc j} = cong suc maxCommutative
maxCommutative {0}     {_}     = refl
maxCommutative {_}     {0}     = refl

leDropMaxRight (i, j, k : ℕ) : (i `max` j ≤ k) -> i ≤ k
leDropMaxRight {0}     {_}             _        = leZ
leDropMaxRight {_}     {0}             le       = le
leDropMaxRight {suc i} {suc j} {suc k} (leS le) = leS (leDropMaxRight le)

leDropMaxLeft (i, j, k : ℕ) : (i `max` j ≤ k) -> j ≤ k
leDropMaxLeft le = leDropMaxRight (castLE maxCommutative refl le)

------------------------------------------------------------------------

-- Calculates n - m.

minus (m, n : ℕ) : m ≤ n -> ℕ
minus {_} {n} leZ        = n
minus         (leS mLeN) = minus mLeN

subtractMinus (m, n : ℕ) : (le : m ≤ n) -> n ∸ m ≡ minus le
subtractMinus leZ      = refl
subtractMinus (leS le) = subtractMinus le

diffMinus : (m, n : ℕ) -> (le : m ≤ n) -> diff m n ≡ minus le
diffMinus 0       n       leZ      = refl
diffMinus (suc m) (suc n) (leS le) = diffMinus m n le

------------------------------------------------------------------------

minusDecreasing (m, n : ℕ) : (le : m ≤ n) -> minus le ≤ n
minusDecreasing leZ      = reflexive
minusDecreasing (leS le) = transitive (minusDecreasing le) sucIncreasing

-- j ≤ i ≤ j + k ⇒ i - j ≤ k

prop_minus_plus (i, j, k : ℕ)
  : i ≤ j + k -> (le : j ≤ i) -> minus le ≤ k
prop_minus_plus {i}     {0}     {k} le'       leZ      = le'
prop_minus_plus {suc i} {suc j} {k} (leS le') (leS le) =
  prop_minus_plus le' le

-- i ≤ j ⇒ i + (j - i) ≡ j

prop_plus_minus (i, j : ℕ)
  : (le : i ≤ j) -> i + minus le ≡ j
prop_plus_minus leZ      = refl
prop_plus_minus (leS le) = cong suc $ prop_plus_minus le

------------------------------------------------------------------------

-- Gives the left answer if possible.

(≤?≥) : (m, n : ℕ) -> m ≤ n ∔ n ≤ m
m ≤?≥ n = mapSum id (transitive sucIncreasing) $ m ≤?> n
