------------------------------------------------------------------------
-- Soundness proof
------------------------------------------------------------------------

-- The thunked operational semantics is sound with respect to the
-- simple one.

module OperationalSemantics.Sound where

open import MainResults
open import Data.Nat
open import Kind
open import TypeSystem
open import Erasure
open import OperationalSemantics.Binding
open Heap↦
open Heap↦Erasure
open import Substitution
open VarSubst
open import Substitution.Lemmas
open import Value
open import OperationalSemantics.Simple  as S
open import OperationalSemantics.Thunked as T
open import OperationalSemantics.GPay
open import OperationalSemantics.GPay.Lemmas
open import OperationalSemantics.Casts
open import Erasure.Lemmas
open import Erasure.Solver
open import Relation.Binary.PropositionalEquality
open import Relation.Binary.HeterogeneousEquality
open ≡-Reasoning

------------------------------------------------------------------------
-- The main soundness proof

mutual

  sound : Soundness

  -- Values.

  sound (T.eval-λ  {Γ = Γ}) = cast S.eval-λ  (substsEqual (⌜id⌝-lemma Γ))
  sound (T.eval-λ² {Γ = Γ}) = cast S.eval-λ² (substsEqual (⌜id⌝-lemma Γ))
  sound (T.eval-z  {Γ = Γ}) = cast S.eval-z  (substsEqual (⌜id⌝-lemma Γ))
  sound (T.eval-s  {Γ = Γ}) = cast S.eval-s  (substsEqual (⌜id⌝-lemma Γ))

  sound (T.eval-natrec {Γ = Γ} {σ = σ} {n = n} {e₂ = e₂}) =
    cast S.eval-natrec
         (resultsEqual (⌜wk⌝-lemma Γ (Thunk n σ)) ≡-refl lemma)
    where
    lemma =
      prove
        (|natrec⇓| (lift vz) (|⌜ lift e₂ ⌝| |/⊢| |wk| |↑| |↑|))
        (|natrec⇓| (lift vz)
                   |⌜ lift e₂ |/⊢|
                      _|↑| {σ = Thunk (suc n) σ}
                           (_|↑| {σ = Nat}
                                 (|wk| {Γ = Γ} {σ = Thunk n σ})) ⌝|)
        ≡-refl

  sound (T.eval-, {Γ = Γ} {σ₁ = σ₁} {σ₂ = σ₂} {Σ = Σ}
                  {e₁ = e₁} {e₂ = e₂}) =
    cast S.eval-, (resultsEqual lemma₁ lemma₂ ≡-refl)
    where
    lemma₁ =
      prove (|wk| {Γ = ⌜ Γ ⌝▻} {σ = ⌜ σ₁ ⌝⋆} |∘ˢ| |wk| {σ = ⌜ σ₂ ⌝⋆})
            (|⌜ |wk| {Γ = Γ} {σ = σ₁} |∘ˢ| |wk| {σ = σ₂} ⌝⇒|)
            ≡-refl

    lemma₂ =
      prove (|⌜ liftʰ Σ ⌝ʰ| |▷ʰ| |↦| |⌜ lift e₁ ⌝| |▷ʰ|
               |↦| |⌜ lift e₂ ⌝| |/⊢| |wk| {σ = ⌜ σ₁ ⌝⋆})
            (|⌜ liftʰ Σ |▷ʰ| |↦| lift e₁ |▷ʰ|
               |↦| lift e₂ |/⊢| |wk| {σ = σ₁} ⌝ʰ|)
            ≡-refl

  sound (T.eval-uncurry {Γ = Γ}) =
    cast S.eval-uncurry (substsEqual (⌜id⌝-lemma Γ))

  -- Variables.

  sound (T.eval-vz {Γ = Γ} {Δ = Δ} {τ = τ} {ms = ms} {n = n} {ρ = ρ}
                   {Σ₁ = Σ₁} {Σ₂ = Σ₂} {e₁ = e₁} {v = v} d) =
    cast (S.eval-vz (sound d)) (record
      { Γ-eq  = ≡-cong (_▻_ ⌜ Γ ⌝▻) lem
      ; τ-eq  = lem
      ; Σ₁-eq = helper₁ lem
      ; t-eq  = helper₂ lem
      ; Δ-eq  = ≡-cong (_▻_ ⌜ Δ ⌝▻) lem
      ; ρ-eq  = helper₃ τ lem
      ; Σ₂-eq = helper₄ (Thunks ms τ) lem lemma₁
      ; v-eq  = helper₅ lem lemma₂
      })
    where
    lem = Thunks-lemma ms

    helper₁ :  forall {σ} (eq : _ ≡ σ)
            -> ⌜ Σ₁ ⌝ʰ ▷ʰ ↦ ⌜ e₁ ⌝ ≅
               ⌜ Σ₁ ⌝ʰ ▷ʰ ↦ ≡-subst (_⊢_ ⌜ Γ ⌝▻) eq ⌜ e₁ ⌝
    helper₁ ≡-refl = ≅-refl

    helper₂ : forall {Γ} {σ₁ σ₂ : Ty simple}
            -> σ₁ ≡ σ₂
            -> var (vz {Γ = Γ} {σ = σ₁}) ≅ var (vz {Γ = Γ} {σ = σ₂})
    helper₂ ≡-refl = ≅-refl

    helper₃ :  forall {σ₁} σ₂
            -> σ₁ ≡ ⌜ σ₂ ⌝⋆ -> _↑ {σ = σ₁} ⌜ ρ ⌝⇒ ≅ ⌜ _↑ {σ = σ₂} ρ ⌝⇒
    helper₃ σ₂ ≡-refl = ≡-to-≅ (⌜↑⌝-lemma σ₂ ρ)

    lemma₁ : valToTm (≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem ⌜ v ⌝⇓) ≡
             ⌜ valToTm (drop-return⋆⇓ ms v) ⌝
    lemma₁ = prove
      (|valToTm| (|≡-subst⇓| lem |⌜ lift v ⌝⇓|))
      (|⌜ |valToTm| (|drop-return⋆⇓| ms (lift v)) ⌝|)
      ≡-refl

    helper₄
      :  forall σ₁ {σ₂} {v₁ : ⌜ Δ ⌝▻ ⊢⇓ ⌜ σ₁ ⌝⋆} {t₂} (eq : ⌜ σ₁ ⌝⋆ ≡ σ₂)
      -> valToTm (≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) eq v₁) ≡ t₂
      -> ⌜ Σ₂ ⌝ʰ ▷ʰ ↦ valToTm v₁ ≅ ⌜ Σ₂ ⌝ʰ ▷ʰ ↦ t₂
    helper₄ _ ≡-refl ≡-refl = ≅-refl

    lemma₂ : ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem ⌜ v ⌝⇓ /⊢⇓ wk ≡
             ⌜ drop-return⋆⇓ ms v /⊢⇓ wk {σ = τ} ⌝⇓
    lemma₂ = prove (|≡-subst⇓| lem |⌜ lift v ⌝⇓| |/⊢⇓| |wk|)
                   |⌜ |drop-return⋆⇓| ms (lift v) |/⊢⇓| |wk| {σ = τ} ⌝⇓|
                   ≡-refl

    helper₅
      :  forall {Γ : Ctxt simple} {σ₁ σ₂} (eq : σ₁ ≡ σ₂)
                {v₁ : Γ ⊢⇓ σ₁} {v₂}
      -> ≡-subst (_⊢⇓_ _) eq v₁ /⊢⇓ wk {σ = σ₂} ≡ v₂
      -> v₁ /⊢⇓ wk {σ = σ₁} ≅ v₂
    helper₅ ≡-refl ≡-refl = ≅-refl

  sound (T.eval-vs {σ = σ} {ρ = ρ} {Σ₂ = Σ₂} {b = b} {v = v} d) =
    cast (S.eval-vs (sound d))
         (resultsEqual (⌜↑⌝-lemma σ ρ) lemma₁ lemma₂)
    where
    lemma₁ =
      prove (|⌜ liftʰ Σ₂ ⌝ʰ| |▷ʰ| |⌜ lift b ⌝↦| |/↦| |⌜ lift⇒ ρ ⌝⇒|)
            (|⌜ liftʰ Σ₂ |▷ʰ| lift b |/↦| lift⇒ ρ ⌝ʰ|)
            ≡-refl

    lemma₂ =
      prove (|⌜ lift v ⌝⇓| |/⊢⇓| |wk|)
            (|⌜ lift v |/⊢⇓| |wk| {σ = σ} ⌝⇓|)
            ≡-refl

  sound (T.eval-gwait∋ {Γ = Γ} {Δ = Δ} {τ = τ} {m₁ = m₁} {m₂ = m₂}
                       {C = C} {x = x} {v = v} d) =
    cast (sound d) (record
      { Γ-eq  = ≡-refl
      ; τ-eq  = lem
      ; Σ₁-eq = ≅-refl
      ; t-eq  = helper₁ lem
      ; Δ-eq  = ≡-refl
      ; ρ-eq  = ≅-refl
      ; Σ₂-eq = ≅-refl
      ; v-eq  = helper₂ lem lemma
      })
    where
    lem = erase-lemma {m = m₁} {n = m₂ + m₁} {τ = τ} C

    helper₁ :  forall {σ} (eq : _ ≡ σ)
            -> var ⌜ x ⌝∋ ≅ var (≡-subst (\τ' -> ⌜ Γ ⌝▻ ∋ τ') eq ⌜ x ⌝∋)
    helper₁ ≡-refl = ≅-refl

    lemma : ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem ⌜ v ⌝⇓ ≡ ⌜ gwait⇓ C m₂ v ⌝⇓
    lemma = prove (|≡-subst⇓| lem |⌜ lift v ⌝⇓|)
                  |⌜ |gwait⇓| C m₂ (lift v) ⌝⇓|
                  ≡-refl

    helper₂ :  forall {σ} (eq : _ ≡ σ) {v'}
            -> ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) eq ⌜ v ⌝⇓ ≡ v'
            -> ⌜ v ⌝⇓ ≅ v'
    helper₂ ≡-refl ≡-refl = ≅-refl

  -- Primitives.

  sound (T.eval-return d)  = sound d
  sound (T.eval-√ d)       = sound d
  sound (T.eval-pay d)     = sound d

  sound (T.eval-gpay {Γ = Γ} {Δ = Δ} {τ = τ} {m₁ = m₁} {m₂ = m₂}
                     {Σ₁ = Σ₁} {Σ₂ = Σ₂} {C = C}
                     {ρ₁ = ρ₁} {e = e} {v = v} d) =
    cast (sound d) (record
      { Γ-eq  = ≡-refl
      ; τ-eq  = lem
      ; Σ₁-eq = ≅-refl
      ; t-eq  = helper₁ lem
      ; Δ-eq  = Δ-eq
      ; ρ-eq  = helper₂ (_⇒_ _) Δ-eq (∘-ρ-lemma rs ρ₁)
      ; Σ₂-eq = helper₂ Heap Δ-eq (GPaySound.Σ₂-eq rs)
      ; v-eq  = helper₃ _⊢⇓_ lem Δ-eq (GPaySound.v₂-eq rs)
      })
    where
    lem  = erase-lemma {m = m₁} {n = m₁ ∸ m₂} {τ = τ} C
    ρ₂   = GPay.ρ (gpayResult C m₂ Σ₂ v)
    rs   = gpayResult-sound C m₂ Σ₂ v
    Δ-eq = GPaySound.Δ-eq rs

    helper₁ :  forall {σ} (eq : _ ≡ σ)
            -> ⌜ e ⌝ ≅ ≡-subst (\τ' -> ⌜ Γ ⌝▻ ⊢ τ') eq ⌜ e ⌝
    helper₁ ≡-refl = ≅-refl

    helper₂ :  forall {a : Set} (P : a -> Set) {u v} (eq : u ≡ v) {x y}
            -> ≡-subst P eq x ≡ y -> x ≅ y
    helper₂ _ ≡-refl ≡-refl = ≅-refl

    helper₃
      :  forall {a b : Set} (P : a -> b -> Set)
                {u v} (eq₁ : u ≡ v) {i j} (eq₂ : i ≡ j) {x y}
      -> ≡-subst (\x -> P x v) eq₂ (≡-subst (P i) eq₁ x) ≡ y -> x ≅ y
    helper₃ _ ≡-refl ≡-refl ≡-refl = ≅-refl

  sound (T.eval->>= {σ = σ} {m₁ = m₁} {n₂ = n₂} {Σ₂ = Σ₂} {Σ₃ = Σ₃}
                    {ρ₁ = ρ₁} {ρ₂ = ρ₂}
                    {e₁ = e₁} {v = v} {v₂ = v₂} d₁ d₂) =
    cast (S.eval-· (sound d₁) (cast• (sound• d₂) lemma₁ ≡-refl))
         (substsEqual lemma₂)
    where
    lemma₁ =
      prove (|⌜ liftʰ Σ₂ |▷ʰ| |↦ⁿ| lift e₁ |/⊢| lift⇒ ρ₁ ⌝ʰ|)
            (|⌜ liftʰ Σ₂ ⌝ʰ| |▷ʰ| |↦| |⌜ lift e₁ ⌝| |/⊢| |⌜ lift⇒ ρ₁ ⌝⇒|)
            ≡-refl

    lemma₂ =
      prove (|wk⇒| |⌜ lift⇒ ρ₁ ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂ ⌝⇒|)
            (|⌜ |wk⇒| (lift⇒ ρ₁) |∘ˢ| lift⇒ ρ₂ ⌝⇒|)
            ≡-refl

  sound (T.eval-gwait {Γ = Γ} {Δ = Δ} {τ = τ} {m₁ = m₁} {m₂ = m₂}
                      {C = C} {e = e} {v = v} d) =
    cast (sound d) (record
      { Γ-eq  = ≡-refl
      ; τ-eq  = lem
      ; Σ₁-eq = ≅-refl
      ; t-eq  = helper₁ lem
      ; Δ-eq  = ≡-refl
      ; ρ-eq  = ≅-refl
      ; Σ₂-eq = ≅-refl
      ; v-eq  = helper₂ lem lemma
      })
    where
    lem = erase-lemma {m = m₁} {n = m₂ + m₁} {τ = τ} C

    helper₁ :  forall {σ} (eq : _ ≡ σ)
            -> ⌜ e ⌝ ≅ ≡-subst (\τ' -> ⌜ Γ ⌝▻ ⊢ τ') eq ⌜ e ⌝
    helper₁ ≡-refl = ≅-refl

    lemma : ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem ⌜ v ⌝⇓ ≡ ⌜ gwait⇓ C m₂ v ⌝⇓
    lemma = prove (|≡-subst⇓| lem |⌜ lift v ⌝⇓|)
                  |⌜ |gwait⇓| C m₂ (lift v) ⌝⇓|
                  ≡-refl

    helper₂ :  forall {σ} (eq : _ ≡ σ) {v'}
            -> ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) eq ⌜ v ⌝⇓ ≡ v'
            -> ⌜ v ⌝⇓ ≅ v'
    helper₂ ≡-refl ≡-refl = ≅-refl

  -- Fix.

  sound (T.eval-fix {Γ = Γ} {σ = σ} {m = m} {ρ = ρ} d) =
    cast (S.eval-fix (sound d)) (substsEqual lemma)
    where
    lemma =
      prove (|wk| |∘ˢ| |⌜ lift⇒ ρ ⌝⇒|)
            (|⌜ |wk| |∘ˢ| lift⇒ ρ ⌝⇒|)
            ≡-refl

  -- Application.

  sound (T.eval-· {σ = σ} {n₂ = n₂} {Σ₂ = Σ₂} {Σ₃ = Σ₃}
                  {ρ₁ = ρ₁} {ρ₂ = ρ₂}
                  {e₂ = e₂} {v = v} {v₁ = v₁} d₁ d₂) =
    cast (S.eval-· (sound d₁) (cast• (sound• d₂) lemma₁ ≡-refl))
         (substsEqual lemma₂)
    where
    lemma₁ =
      prove (|⌜ liftʰ Σ₂ |▷ʰ| |↦| lift e₂ |/⊢| lift⇒ ρ₁ ⌝ʰ|)
            (|⌜ liftʰ Σ₂ ⌝ʰ| |▷ʰ| |↦| |⌜ lift e₂ ⌝| |/⊢| |⌜ lift⇒ ρ₁ ⌝⇒|)
            ≡-refl

    lemma₂ =
      prove (|wk⇒| |⌜ lift⇒ ρ₁ ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂ ⌝⇒|)
            (|⌜ |wk⇒| (lift⇒ ρ₁) |∘ˢ| lift⇒ ρ₂ ⌝⇒|)
            ≡-refl

  sound• :  forall {Γ Δ σ τ ρ Σ₁ Σ₂ n b}
                   {v₁ : Γ ⊢⇓ σ → τ} {v : Δ ⊢⇓ τ}
         -> Σ₁ ▷ʰ b ∣ v₁ •⇊ ρ ⋉ Σ₂ ∣ v √ n
         -> ⌜ Σ₁ ▷ʰ b ⌝ʰ ∣ ⌜ v₁ ⌝⇓ •⇓ ⌜ ρ ⌝⇒ ⋉ ⌜ Σ₂ ⌝ʰ ∣ ⌜ v ⌝⇓ √ n
  sound• (T.eval-·-s {Γ = Γ}) =
    cast• S.eval-·-s ≡-refl (⌜id⌝-lemma (Γ ▻ Nat))
  sound• (T.eval-·-λ d) =
    S.eval-·-λ (sound d)
  sound• (T.eval-·-λ² {Γ = Γ} {σ₁ = σ₁}) =
    cast• S.eval-·-λ² ≡-refl (⌜id⌝-lemma (Γ ▻ σ₁))

  -- Natrec zero.

  sound• (T.eval-·-natrec-zero
             {Γ₁ = Γ₁} {ρ₁ = ρ₁} {ρ₂ = ρ₂} {x₁¹ = x₁¹} d₁ d₂) =
    cast• (S.eval-·-natrec-zero (sound d₁)
                                (cast (sound d₂) (termsEqual lemma)))
          ≡-refl
          (⌜∘⌝-lemma ρ₁ ρ₂)
    where
    lemma =
      prove (|⌜ |var| (lift x₁¹) |/⊢| |wk| |∘ˢ| lift⇒ ρ₁ ⌝|)
            (|var| |⌜ lift x₁¹ ⌝∋| |/⊢| |wk| |∘ˢ| |⌜ lift⇒ ρ₁ ⌝⇒|)
            ≡-refl

  -- Natrec suc.

  sound• (T.eval-·-natrec-suc {ρ₁ = ρ₁} {ρ₂ = ρ₂} {Σ₂ = Σ₂}
                              {x₁¹ = x₁¹} {e₁² = e₁²} {x₂ = x₂} d₁ d₂) =
    cast• (S.eval-·-natrec-suc
             (sound d₁)
             (cast (sound d₂) (inputsEqual lemma₂ lemma₃)))
          ≡-refl
          lemma₁
    where
    lemma₁ =
      prove (|wk⇒| |⌜ lift⇒ ρ₁ ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂ ⌝⇒|)
            (|⌜ |wk⇒| (lift⇒ ρ₁) |∘ˢ| lift⇒ ρ₂ ⌝⇒|)
            ≡-refl

    lemma₂ =
      prove (|⌜ liftʰ Σ₂ |▷ʰ|
             |↦| (|natrec| (|var| (lift x₁¹)) (lift e₁²) |/⊢|
                  |wk| |∘ˢ| lift⇒ ρ₁) |·|
                 |var| (lift x₂) ⌝ʰ|)
            (|⌜ liftʰ Σ₂ ⌝ʰ| |▷ʰ|
             |↦| ((|natrec| (|var| |⌜ lift x₁¹ ⌝∋|) |⌜ lift e₁² ⌝| |/⊢|
                   |wk| |∘ˢ| |⌜ lift⇒ ρ₁ ⌝⇒|) |·|
                  |var| |⌜ lift x₂ ⌝∋|))
            ≡-refl

    lemma₃ =
      prove (|⌜ lift e₁² |/⊢| (|wk| |∘ˢ| lift⇒ ρ₁) |↑| |↑| |/⊢|
             |sub| (lift x₂) |↑| ⌝|)
            (|⌜ lift e₁² ⌝| |/⊢| (|wk| |∘ˢ| |⌜ lift⇒ ρ₁ ⌝⇒|) |↑| |↑| |/⊢|
             |sub| |⌜ lift x₂ ⌝∋| |↑|)
            ≡-refl

  -- Uncurry.

  sound• (T.eval-·-uncurry
             {Γ₁ = Γ₁} {σ₁ = σ₁} {σ₂ = σ₂} {ρ₁ = ρ₁} {ρ₂ = ρ₂}
             {e₁ = e₁} {x₂ = x₂} {y₂ = y₂} d₁ d₂) =
    cast• (S.eval-·-uncurry (sound d₁)
                            (cast (sound d₂) (termsEqual lemma)))
          ≡-refl
          (⌜∘⌝-lemma ρ₁ ρ₂)
    where
    lemma =
      prove (|⌜ lift e₁ |/⊢| (|wk| |∘ˢ| lift⇒ ρ₁) |↑| |↑| |/⊢|
             |sub| (lift x₂) |↑| |/⊢| |sub| (lift y₂) ⌝|)
            (|⌜ lift e₁ ⌝| |/⊢| (|wk| |∘ˢ| |⌜ lift⇒ ρ₁ ⌝⇒|) |↑| |↑| |/⊢|
             |sub| |⌜ lift x₂ ⌝∋| |↑| |/⊢| |sub| |⌜ lift y₂ ⌝∋|)
            ≡-refl
